#include "sivia.h"

//--------------------------------------------------------------------
// Parameters for the contractor
//--------------------------------------------------------------------
extern double xv,yv,theta;
extern double d1,d2,delta;
extern vector<double> murs_xa,murs_ya,murs_xb,murs_yb;
//--------------------------------------------------------------------


//--------------------------------------------------------------------
// Constructor / Destructor
//--------------------------------------------------------------------
Sivia::Sivia(mSimpleDrawer *R, double epsilon)
{
    this->R = R;
    this->epsilon = 0.01;
    H0 = NULL;
}
//--------------------------------------------------------------------
Sivia::~Sivia(){
    if(H0 != NULL) delete H0;
}

//--------------------------------------------------------------------
// Entry point from MainWindow
//--------------------------------------------------------------------
void Sivia::runSivia(Separator C, box X0, double epsilon)
{
    this->epsilon = epsilon;
    // remise à niveau des statistiques
    clearStats();
    if(H0 != NULL) delete H0;

    QElapsedTimer tmp;

    H0 = new HyperBox(X0);
    tmp.start();
    SIVIA(C,H0);
    exec_time = tmp.elapsed();
    drawTree(H0);
    printStats();
}


//--------------------------------------------------------------------
// Statistics
//--------------------------------------------------------------------
void Sivia::clearStats(){
    nb_box = 0; nb_sep_call = 0; nb_emptyBox = 0; casNonTraite = 0; nb_yellow = 0;
    nbInBox = 0; nbOutBox = 0; nbYellowBox  = 0; nbBoxTotal = 0; remainingEmpty = 0;
}
//--------------------------------------------------------------------
void Sivia::printStats(){
    //int nbBoxTotal = nbInBox + nbOutBox + nbYellowBox;
    qDebug() << "======================= STATISTIQUES ===============";
    qDebug() << "Temps exec:" << exec_time/1000.0;
    qDebug() << "Number of bisection:" << nb_box << "\tNumber of empty boxes generated:" << nb_emptyBox;
    qDebug() << "Number of separator calls: "<< nb_sep_call << "\tNumber of Yellow boxes generated:" << nb_yellow;
    qDebug() << "Number of boxes:" << nbBoxTotal << "\tEmpty boxes remaining:" << remainingEmpty;
    qDebug() << "Number of boxes In/Out:" << nbInBox << "/" << nbOutBox << "\tNumber of Yellow boxes:" << nbYellowBox;

    if(casNonTraite) qDebug() << "Number of errors:" << casNonTraite;
}





// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ------------------------------   Atomic Contractor   ----------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------


//----------------------------------------------------------------------------------------
// Map contractor
//----------------------------------------------------------------------------------------
double argument(double mx, double my, double xa, double ya, double xb, double yb ){
    double ma_x = xa - mx;  double ma_y = ya - my;
    double mb_x = xb - mx;  double mb_y = yb - my;
    double norm_ma = hypot(ma_x, ma_y); double norm_mb = hypot(mb_x, mb_y);
    double cos_theta = (ma_x*mb_x + ma_y*mb_y) / (norm_ma*norm_mb);
    double sin_theta = (ma_x*mb_y - ma_y*mb_x) / (norm_ma*norm_mb);
    return atan2(sin_theta,cos_theta);
}

// Retourne true si la boite est à l'extérieur du polygone
bool check(box &X){
    double theta = 0;
    for(uint i = 0; i < murs_xa.size(); i++){
        theta += argument(Center(X[1]),Center(X[2]),murs_xa[i],murs_ya[i],murs_xb[i],murs_yb[i]);
    }
    if(theta > 1) return true;
    return false;
}

void ContractPointsOutPoly(box &X, bool outer){
    box Xold(X);
    vector<box>*dif;
    vector<box> LXin, LXout;
    CPointInSegments(X[1],X[2],murs_xa,murs_ya,murs_xb,murs_yb);
    dif = diff(Xold,X);
    for(int i = 0; i  < dif->size(); i++){
        box Xt = dif->at(i);
        if(check(Xt) == false){
            LXout.push_back(Xt);
        } else {
            LXin.push_back(Xt);
        }
    }
    delete dif;
    if(outer == true){
        if(LXin.size() > 0)
            X = Union(X, Union(LXin));
    } else {
        if(LXout.size() > 0)
            X = Union(X, Union(LXout));
    }
}

//----------------------------------------------------------------------------------------
// Others contractor
//----------------------------------------------------------------------------------------
void CRotTransRotTrans(interval& qx, interval &qy, interval& a1, interval &a2, interval &px, interval &py, double d1, double d2){
    interval qx_tmp(-oo, oo); interval qy_tmp(-oo, oo);
    interval id1(d1 - 0.001, d1 + 0.001); interval id2(d2 - 0.001, d2 + 0.001);
    interval a3 = a1 + a2;
    CPoseTrans(qx_tmp,qy_tmp,id1,px,py,a1);
    CPoseTrans(qx,qy,id2,qx_tmp,qy_tmp,a3);
    Cplus(a3,a1,a2,-1);
    CPoseTrans(qx_tmp,qy_tmp,id1,px,py,a1);
}

//----------------------------------------------------------------------------------------
void CRotTransRotTransInPoly(interval &a1, interval &a2, double d1, double d2, bool outer){
    interval qx(-oo, oo); interval qy(-oo, oo);
    box Xq(interval(-oo, oo),2);
    interval px0(xv - 1e-5, xv + 1e-5);     interval py0(yv - 1e-5, yv + 1e-5);

    CRotTransRotTrans(Xq[1],Xq[2],a1,a2,px0,py0,d1,d2);
    ContractPointsOutPoly(Xq,outer);
    CRotTransRotTrans(Xq[1],Xq[2],a1,a2,px0,py0,d1,d2);

}
//----------------------------------------------------------------------------------------
void CRotTransRotPoly(interval &px, interval &py, interval d, double d1, bool outer){
    interval pxm(px), pym(py);
    interval pxp(px), pyp(py);
    interval pxm1(px), pym1(py);
    interval pxp1(px), pyp1(py);

    CRotTransRotTransInPoly(pxm,pym, d1, d.inf, outer);
    CRotTransRotTransInPoly(pxp,pyp, d1, d.sup, !outer);

    if(outer == true){
        px = Inter(px, Inter(pxm, pxp));
        py = Inter(py, Inter(pym, pyp));

    } else {
        px = Union(pxm, pxp);
        py = Union(pym, pyp);
    }
}


void Sivia::separateur(box &X, bool outer){
    interval d(d2, d2+delta);
    CRotTransRotPoly(X[1],X[2],d,d1,outer);
}


bool checkContract(void (*C)(box& X, bool outer), box X,bool outer){
    vector<box>*dif;
    box Xold(X);
    C(X,outer);
    dif = diff(Xold,X);
    return (dif->size() > 0);
}

// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// -------------------------------------   Fix Point   ----------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------

bool IsContracted(box &X, box &Y){
    if (Y.IsEmpty()) return false;
    if (X.IsEmpty()) return true;
    bool b=false;
    for (int k=1; k<=Size(X); k++){
        bool b1 = (abs(X[k].inf - Y[k].inf) > 1e-7);
        bool b2 = (abs(X[k].sup - Y[k].sup) > 1e-7);
        if( b1 || b2){
            return true;
        } else {
            X = Y; // on rejete la contraction
        }
    }
    return false;
}

// ---------------------------------------------------------------------------------------------------
void removeOverlapedPart(vector<box> *dif, box &X, box &ref){
    for(uint i = 0; i < dif->size(); i++){
        box t = Inter(dif->at(i),ref);
        if(t.IsEmpty()){
            X = dif->at(i);
        }
    }
}

// ---------------------------------------------------------------------------------------------------
enum{NoIncl = 0, XinInclXout, XoutInclXin};
bool generateCompl(box &X0, box &Xin, box &cIn, box &Xout, box &cOut){
    vector<box> *difIn = diff(X0,Xin);
    vector<box> *difOut = diff(X0,Xout);
    int inclusion = Subset(Xin, Xout) + (Subset(Xout, Xin) << 1);
    switch(inclusion){
    case NoIncl:
        cIn = difIn->at(0); cOut = difOut->at(0);
        break;
    case XinInclXout:
        cOut = difOut->at(0);
        removeOverlapedPart(difIn,cIn, cOut);
        break;
    case XoutInclXin:
        cIn = difIn->at(0);
        removeOverlapedPart(difOut,cOut, cIn);
        break;
    default:
        qDebug() << "erreur on ne devrait pas etre la !!!";
        break;
    }
    delete difIn; delete difOut;
    box t = cIn & cOut;
    if(t.IsEmpty() == false) qDebug() << "cIn & Cout non vide !!!" << Width(t) << cIn << cOut << X0 << Xin << Xout;
    return false;
}

// ---------------------------------------------------------------------------------------------------
void identifyInOut(vector<box> *difX, box &X, box &cIn, box &cOut, box &Xin, box &Xout){

    bool b1 = (Inter(cIn,difX->at(0)).IsEmpty() == false);
    bool b2 = (Inter(cIn,difX->at(1)).IsEmpty() == false);
    bool b3 = (Inter(cOut,difX->at(0)).IsEmpty() == false);
    bool b4 = (Inter(cOut,difX->at(1)).IsEmpty() == false);

    bool t = (b1^b3) && (b2^b4) && (b1^b3) && (b3^b4);
    if( t == false){
        qDebug() << "erreur combinaison cIn cOut";
        return;
    }

    box d1 = difX->at(0); box d2 = difX->at(1);
    Xout = (b1) ? Union(X,d1) : Union(X,d2);
    Xin = (b3) ?  Union(X,d1) : Union(X,d2);
}

// ---------------------------------------------------------------------------------------------------
// Cette fonction permet d'identifier les parties qui sont retirées par le contracteur et les retourne dans Xin et Xout
// Paramètres :
//  - input
//      <X0>    : Boite initiale
//      <X>     : Résultat de la contraction point fixe
//      <LXin>  : contient toutes les boites contractées par Cin
//      <LXout>  : contient toutes les boites contractées par Cout
//  - output
//      <Xin>   : contiendra la partie enlevée par le contracteur intérieur union X
//      <Xout>  : contiendra la partie enlevée par le contracteur extérieur union X
//
// ------------------------------------------------------------------------
// Fonctionnement
//      on part du principe que si LXin et LXout contiennent des éléments, alors la contraction
//      a été faite suivant une et une seule direction. il faut donc identifier quelles sont les parties
//      qui ont été enlevé par Cin et Cout
//      Si LXin est vide et LXout non => seulement une contraction par Cout
//      Si LXout est vide et LXin non => seulement une contraction par Cin
//      Si LXout et LXin sont vide => aucune contraction.
//
enum{NoContraction = 0, CoutOnly, CinOnly, CinAndCout};
void indentification(box &X0, box &X, box &Xin, box &Xout, vector<box> &LXin, vector<box> &LXout,int cmp){
    vector<box> *difX;
    box cIn(X.dim),cOut(X.dim);
    int contractedState = (!LXin.empty() << 1) + !LXout.empty();
    switch(contractedState){
    case NoContraction:
        if (!(X==X0))
            qDebug() << "erreur pas de contraction  mais X != X0" << LXin.empty() << LXout.empty() << cmp;
        Xin = X0; Xout = X0;
        break;
    case CoutOnly: // Contraction uniquement par Cout
        Xout = X; Xin =  X0;
        break;
    case CinOnly:   // Conctraction uniquement par Cin
        Xin = X; Xout = X0;
        break;
    case CinAndCout: // Contraction mixte par Cin et Cout
        difX = diff(X0,X);
        if(difX->size() != 2) { qDebug() << "erreur diff(X0 X)" << difX->size() << X0 << X;}
        generateCompl(X0, LXin[0], cIn, LXout[0], cOut);
        identifyInOut(difX, X, cIn, cOut, Xin, Xout);
        delete difX;
        break;
    default:
        qDebug() << "On ne devrait pas être ici !!!!" ;
    }
}

// ---------------------------------------------------------------------------------------------------
box Sivia::fixPoint(Separator S, box &X0, box &Xin, box &Xout){

    // Debut de la partie point fixe
    bool sortie=false;

    box X(X0); Xin = X0; Xout = X0;

    //Xin = box(); Xout = box();
    vector<box> LXout;
    vector<box> LXin;
    int cmp = 0;
    while (!sortie)
    {
        box Xold(X); box Xin_tmp(X); box Xout_tmp(X);
        nb_sep_call++;
        (this->*S)(Xout_tmp, true);
        (this->*S)(Xin_tmp, false);
        if(IsContracted(Xout_tmp,Xold)) LXout.push_back(Xout_tmp);
        if(IsContracted(Xin_tmp,Xold)) LXin.push_back(Xin_tmp);
        X = Inter(Xin_tmp, Xout_tmp);
        if (X.IsEmpty())      sortie=true;
        if (decrease(Xold,X)<1e-3) sortie=true;
        cmp++;
    }
    // Fin de la partie point fixe.
    indentification(X0,X,Xin,Xout,LXin,LXout, cmp);
    return X;
}



// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------    SIVIA   ------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------

void Sivia::reduceNode(HyperBox* H0){
    HyperBox* H1= H0->getChild(0);
    HyperBox* H2= H0->getChild(1);

    vector<box>*cXin= diff(H0->X0(),H0->Xin);
    vector<box>*cXout= diff(H0->X0(),H0->Xout);
    H0->Xin =  H1->Xin | H2->Xin | Union(*cXout);
    H0->Xout = H1->Xout | H2->Xout | Union(*cXin);

    int t = (H1->IsEmpty() << 1) + H2->IsEmpty();
    if( t > 0){
        H0->reduce();
    } else {
        H1->Xin = H0->Xout & H0->Xin & H1->Xin;
        H1->Xout = H0->Xout & H0->Xin & H1->Xout;
        H2->Xin = H0->Xout & H0->Xin & H2->Xin;
        H2->Xout = H0->Xout & H0->Xin & H2->Xout;
    }
     delete cXin; delete cXout;
}
//--------------------------------------------------------------------
void Sivia::SIVIA(Separator S,HyperBox *H0)    // Un separateur est donn� comme param�tre : on passe un pointeur sur fonction.
{
    box X0 = H0->X0();
    box X = fixPoint(S,X0, H0->Xin, H0->Xout);

    if(X.IsEmpty()){
        nb_emptyBox++;
        return;
    }
    if (X.Width()<epsilon)
    {
        nb_yellow++;
        return ;
    }

    HyperBox *H1 = new HyperBox(H0), *H2 = new HyperBox(H0);
    H0->HyperBisect(H1,H2);

    nb_box+=1;

    SIVIA(S,H1);
    SIVIA(S,H2);

    reduceNode(H0);
}
//--------------------------------------------------------------------
void Sivia::drawTree(HyperBox *H0){
    list<HyperBox*> L;
    L.push_back(H0);

    while(!L.empty()){
        HyperBox *H = L.back(); L.pop_back();
        box X0 = H->X0();
        nbOutBox += R->DrawBoxDiff(X0, H->Xout,QPen(Qt::gray),QBrush(Qt::lightGray));
        nbInBox += R->DrawBoxDiff(X0, H->Xin,QPen(Qt::black), QBrush(Qt::darkGray));
        box X = H->Xin & H->Xout;
        if(X.IsEmpty()) remainingEmpty++;
        if(X.Width() < epsilon){
            R->DrawBox(X[1].inf,X[1].sup,X[2].inf,X[2].sup,QPen(Qt::white),QBrush(Qt::green));
            nbYellowBox++;
        }
        if(H->getChild(0) != NULL) L.push_back(H->getChild(0));
        if(H->getChild(1) != NULL) L.push_back(H->getChild(1));
        nbBoxTotal++;
    }
}

